// SharedMemory.cpp : Master & Slave
// 2016.05 Kyoung Shin Park 
//

#include <windows.h>
#include <iostream>
#include <string.h>
#include <time.h>

#include "SharedMemory.h"
#include "Semaphore.h"
#include "Mutex.h"

using namespace std;

//#define BUFFER_SIZE 4096

#define NUM_TOTAL_BUFFERS 5
#define DATA_LENGTH 20

// shared memory for buffer data
CSharedMemory shmem("_SHMEM_BUFFER");
const key_t shmBufferSegmentKey = 3782;
HANDLE shmBufferSegmentID = NULL;
int * gBuffers; // global shmem data buffer (5 integers)
// shared memory for reader index
CSharedMemory readerIndexShmem("_SHMEM_READER_COUNTER");
const key_t shmIndexSegmentKey = 3792;
HANDLE shmIndexSegmentID = NULL;
int * gReaderIndex = NULL; // shmem for global reader counter between 2 readers

// semaphore 
KSemaphore emptyBufferSemaphore("_EMPTY_BUFFER"); 
KSemaphore fullBufferSemaphore("_FULL_BUFFER");

// mutex
KMutex readerMutex("_READER");

bool isMaster = false;

int randomVal(int min, int max)
{
	static bool first = true;
	if (first)
	{
		srand(time((time_t)NULL)); //seeding for the first time only!
		first = false;
	}
	return min + rand() % (max - min);
}

void parseCommandLine(int argc, char ** argv)
{
    for (int i = 1; i < argc; i++) {
		if (!strncmp(argv[i], "--", 2)) {
			if (!strncmp(argv[i], "--master", 8)) {
				isMaster = true;
				printf("MASTER\n");
			}
		}
	}
}

void writer()
{
	int gWriterIndex = 0; // global writer counter 
	for (int i = 0; i < DATA_LENGTH; i++)
	{
		Sleep(200); // window Sleep
		emptyBufferSemaphore.lock();
		int value = randomVal(200, 500);
		gBuffers[gWriterIndex] = value;
		printf("writerThread[%d]: buffers[%d] = %d\n", i, gWriterIndex, value);
		gWriterIndex = (gWriterIndex + 1) % NUM_TOTAL_BUFFERS;
		fullBufferSemaphore.unlock();
	}

}

void reader()
{
	for (int i = 0; i < DATA_LENGTH; i++)
	{
		fullBufferSemaphore.lock();
		readerMutex.lock();
		int value = gBuffers[(*gReaderIndex)];
		printf("                readerThread[%d]: buffers[%d] = %d\n", i, (*gReaderIndex), value);
		(*gReaderIndex) = ((*gReaderIndex) + 1) % NUM_TOTAL_BUFFERS;
		readerMutex.unlock();
		emptyBufferSemaphore.unlock();
		int s = randomVal(200, 700);
		Sleep(s); // window Sleep
	}
}

// 1 master (1 writer) & 2 slaves (2 readers)
int main(int argc, char ** argv)
{
	parseCommandLine(argc, argv);
	
	int bufferSize = sizeof(int) * 5;
	if (isMaster) { // master is a writer
		// init & allocate shared memory segment for buffer data
		shmBufferSegmentID = shmem.getShm(shmBufferSegmentKey, bufferSize, S_IRUSR | S_IWUSR | IPC_CREAT); // create shmem buffer data
		printf("shmemBuffer id=%d\n", (int)shmBufferSegmentID);
		// attach the shared memory segment
		gBuffers = (int *) shmem.attachShm(shmBufferSegmentID, NULL, 0); // alloc shmem buffer data
		printf("shared memory attached\n");
		for (int i = 0; i < NUM_TOTAL_BUFFERS; i++) { // initialize to 0
			gBuffers[i] = 0;
		}
		// create semaphore
		emptyBufferSemaphore.create(NUM_TOTAL_BUFFERS, NUM_TOTAL_BUFFERS);
		fullBufferSemaphore.create(0, NUM_TOTAL_BUFFERS);
	}
	else { // slaves are readers
		// get shared memory segment for buffer data
		shmBufferSegmentID = shmem.getShm(shmBufferSegmentKey, bufferSize, S_IRUSR | S_IWUSR); // open shmem buffer data
		printf("shmemBuffer id=%d\n", (int)shmBufferSegmentID);
		// attach the shared memory segment
		gBuffers = (int *)shmem.attachShm(shmBufferSegmentID, NULL, 0); // attach
		printf("shared memory attached\n");
		// get shared memory segment for gReaderIndex counter
		bool indexSegmentCreated = true;
		shmIndexSegmentID = readerIndexShmem.getShm(shmIndexSegmentKey, sizeof(int), S_IRUSR | S_IWUSR | IPC_CREAT); // create shmem index
		if (shmIndexSegmentID == NULL) {
			printf("Could not create shared memory for readerIndexShmem\n");
			exit(-1);
		}
		gReaderIndex = (int *)readerIndexShmem.attachShm(shmIndexSegmentID, NULL, 0); // alloc shmem index
		if (readerIndexShmem.isCreated()) {
			(*gReaderIndex) = 0;
			printf("gReaderIndex=%d\n", (*gReaderIndex));
		}
		printf("shmemIndex id=%d\n", (int)shmIndexSegmentID);
		// reader mutex create/open
		if (!readerMutex.create()) { // try to create Mutex (only 1 reader creates a mutex)
			printf("need to open mutex\n");
			readerMutex.open(); // if one is already created, other readers should open a mutex
		}
		// open semaphore
		emptyBufferSemaphore.open();
		fullBufferSemaphore.open();
	}

	// while loop
	if (isMaster) {
		writer();
	}
	else {
		reader();
	}

	// detach the shared memory segment
	shmem.detachShm(gBuffers);
	readerIndexShmem.detachShm(gReaderIndex);
	// remove the shared memory segment
	shmem.cleanShm(shmBufferSegmentID, IPC_RMID, NULL);
	readerIndexShmem.cleanShm(shmIndexSegmentID, IPC_RMID, NULL);
	printf("shared memory detached & cleanup \n");
	// close semaphore
	emptyBufferSemaphore.close();
	fullBufferSemaphore.close();
	// close mutex
	readerMutex.close();
	printf("close semaphore & mutex.\n");

	return 0;
}

